@@ -19,12 +19,7 @@ You should have received a copy of the GNU General Public License along
 with Bash; see the file COPYING.  If not, write to the Free Software
 Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 
-#if defined (HAVE_STDLIB_H)
-#  include <stdlib.h>
-#else
-#  include "ansi_stdlib.h"
-#endif /* HAVE_STDLIB_H */
-
+#include "../bashansi.h"
 #include "../config.h"
 #include <stdio.h>
 #include <sys/types.h>
@@ -50,7 +45,7 @@ static char *xmalloc (), *xrealloc ();
 extern char *strcpy ();
 #endif /* !__STDC__ && !strcpy */
 
-#define savestring(x) (char *)strcpy (xmalloc (1 + strlen (x)), (x))
+#define savestring(x) strcpy (xmalloc (1 + strlen (x)), (x))
 #define whitespace(c) (((c) == ' ') || ((c) == '\t'))
 
 /* Flag values that builtins can have. */
@@ -83,7 +78,7 @@ char *extern_filename = (char *)NULL;
 /* Here is a structure for manipulating arrays of data. */
 typedef struct {
   int size;		/* Number of slots allocated to array. */
-  int index;		/* Current location in array. */
+  int sindex;		/* Current location in array. */
   int width;		/* Size of each element. */
   int growth_rate;	/* How fast to grow. */
   char **array;		/* The array itself. */
@@ -157,7 +152,7 @@ main (argc, argv)
 	  if (error_directory)
 	    free (error_directory);
 
-	  error_directory = (char *)xmalloc (2 + strlen (argv[arg_index]));
+	  error_directory = xmalloc (2 + strlen (argv[arg_index]));
 	  strcpy (error_directory, argv[arg_index]);
 	  len = strlen (error_directory);
 
@@ -191,7 +186,7 @@ main (argc, argv)
       /* Open the files. */
       if (struct_filename)
 	{
-	  temp_struct_filename = (char *)xmalloc (15);
+	  temp_struct_filename = xmalloc (15);
 	  sprintf (temp_struct_filename, "mk-%d", (int) getpid ());
 	  structfile = fopen (temp_struct_filename, "w");
 
@@ -271,7 +266,7 @@ array_create (width)
 
   array = (ARRAY *)xmalloc (sizeof (ARRAY));
   array->size = 0;
-  array->index = 0;
+  array->sindex = 0;
   array->width = width;
 
   /* Default to increasing size in units of 20. */
@@ -296,12 +291,12 @@ copy_string_array (array)
   copy = array_create (sizeof (char *));
 
   copy->size = array->size;
-  copy->index = array->index;
+  copy->sindex = array->sindex;
   copy->width = array->width;
 
-  copy->array = (char **)xmalloc ((1 + array->index) * sizeof (char *));
+  copy->array = (char **)xmalloc ((1 + array->sindex) * sizeof (char *));
   
-  for (i = 0; i < array->index; i++)
+  for (i = 0; i < array->sindex; i++)
     copy->array[i] = savestring (array->array[i]);
 
   copy->array[i] = (char *)NULL;
@@ -314,17 +309,17 @@ array_add (element, array)
      char *element;
      ARRAY *array;
 {
-  if (array->index + 2 > array->size)
+  if (array->sindex + 2 > array->size)
     array->array = (char **)xrealloc
       (array->array, (array->size += array->growth_rate) * array->width);
 
 #if defined (HAVE_BCOPY)
-  bcopy (&element, &(array->array[array->index]), array->width);
-  array->index++;
-  bzero (&(array->array[array->index]), array->width);
+  bcopy (&element, &(array->array[array->sindex]), array->width);
+  array->sindex++;
+  bzero (&(array->array[array->sindex]), array->width);
 #else
-  array->array[array->index++] = element;
-  array->array[array->index] = (char *)NULL;
+  array->array[array->sindex++] = element;
+  array->array[array->sindex] = (char *)NULL;
 #endif /* !HAVE_BCOPY */
 }
 
@@ -475,6 +470,7 @@ extract_info (filename, structfile, externfile)
 	  if (!handler)
 	    {
 	      line_error (defs, "Unknown directive `%s'", directive);
+	      free (directive);
 	      continue;
 	    }
 	  else
@@ -486,6 +482,7 @@ extract_info (filename, structfile, externfile)
 	      /* Call the directive handler with the FILE, and ARGS. */
 	      (*(handler->function)) (directive, defs, line + j);
 	    }
+	  free (directive);
 	}
       else
 	{
@@ -495,10 +492,16 @@ extract_info (filename, structfile, externfile)
 	    {
 	      if (output_cpp_line_info)
 		{
-		  fprintf (defs->output, "#line %d \"%s%s\"\n",
-			   defs->line_number + 1,
-			   error_directory ? error_directory : "./",
-			   defs->filename);
+		  /* If we're handed an absolute pathname, don't prepend
+		     the directory name. */
+		  if (defs->filename[0] == '/')
+		    fprintf (defs->output, "#line %d \"%s\"\n",
+			     defs->line_number + 1, defs->filename);
+		  else
+		    fprintf (defs->output, "#line %d \"%s%s\"\n",
+			     defs->line_number + 1,
+			     error_directory ? error_directory : "./",
+			     defs->filename);
 		  output_cpp_line_info = 0;
 		}
 
@@ -522,35 +525,48 @@ extract_info (filename, structfile, externfile)
 
 #define free_safely(x) if (x) free (x)
 
+static void
+free_builtin (builtin)
+     BUILTIN_DESC *builtin;
+{
+  register int i;
+
+  free_safely (builtin->name);
+  free_safely (builtin->function);
+  free_safely (builtin->shortdoc);
+  free_safely (builtin->docname);
+
+  if (builtin->longdoc)
+    array_free (builtin->longdoc);
+
+  if (builtin->dependencies)
+    {
+      for (i = 0; builtin->dependencies->array[i]; i++)
+	free (builtin->dependencies->array[i]);
+      array_free (builtin->dependencies);
+    }
+}
+
 /* Free all of the memory allocated to a DEF_FILE. */
 free_defs (defs)
      DEF_FILE *defs;
 {
-  register int i, j;
+  register int i;
   register BUILTIN_DESC *builtin;
 
   if (defs->production)
     free (defs->production);
 
+  if (defs->lines)
+    array_free (defs->lines);
+
   if (defs->builtins)
     {
       for (i = 0; builtin = (BUILTIN_DESC *)defs->builtins->array[i]; i++)
-	{
-	  free_safely (builtin->name);
-	  free_safely (builtin->function);
-	  free_safely (builtin->shortdoc);
-
-	  if (builtin->dependencies)
-	    {
-	      for (j = 0; builtin->dependencies->array[j]; j++)
-		free (builtin->dependencies->array[j]);
-
-	      array_free (builtin->dependencies);
-	    }
-
-	  if (builtin->longdoc)
-	    array_free (builtin->longdoc);
-	}
+        {
+	  free_builtin (builtin);
+	  free (builtin);
+        }
       array_free (defs->builtins);
     }
   free (defs);
@@ -624,7 +640,7 @@ current_builtin (directive, defs)
      DEF_FILE *defs;
 {
   must_be_building (directive, defs);
-  return ((BUILTIN_DESC *)defs->builtins->array[defs->builtins->index - 1]);
+  return ((BUILTIN_DESC *)defs->builtins->array[defs->builtins->sindex - 1]);
 }
 
 /* Add LINE to the long documentation for the current builtin.
@@ -823,9 +839,9 @@ line_error (defs, format, arg1, arg2)
      DEF_FILE *defs;
      char *format, *arg1, *arg2;
 {
-  fprintf (stderr, "%s%s:%d:",
-	   error_directory ? error_directory : "./",
-	   defs->filename, defs->line_number + 1);
+  if (defs->filename[0] != '/')
+    fprintf (stderr, "%s", error_directory ? error_directory : "./");
+  fprintf (stderr, "%s:%d:", defs->filename, defs->line_number + 1);
   fprintf (stderr, format, arg1, arg2);
   fprintf (stderr, "\n");
   fflush (stderr);
@@ -1032,7 +1048,7 @@ write_builtins (defs, structfile, externfile)
     {
       register BUILTIN_DESC *builtin;
 
-      for (i = 0; i < defs->builtins->index; i++)
+      for (i = 0; i < defs->builtins->sindex; i++)
 	{
 	  builtin = (BUILTIN_DESC *)defs->builtins->array[i];
 
@@ -1117,7 +1133,7 @@ write_longdocs (stream, builtins)
   register int i;
   register BUILTIN_DESC *builtin;
 
-  for (i = 0; i < builtins->index; i++)
+  for (i = 0; i < builtins->sindex; i++)
     {
       builtin = (BUILTIN_DESC *)builtins->array[i];
 
