@@ -21,6 +21,10 @@
    675 Mass Ave, Cambridge, MA 02139, USA. */
 #define READLINE_LIBRARY
 
+#if defined (HAVE_CONFIG_H)
+#  include "config.h"
+#endif
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <fcntl.h>
@@ -226,6 +230,13 @@ int rl_ignore_completion_duplicates = 1;
    within a completion entry finder function. */
 int rl_filename_completion_desired = 0;
 
+/* Non-zero means that the results of the matches are to be quoted using
+   double quotes (or an application-specific quoting mechanism) if the
+   filename contains any characters in rl_word_break_chars.  This is
+   ALWAYS non-zero on entry, and can only be changed within a completion
+   entry finder function. */
+int rl_filename_quoting_desired = 1;
+
 /* This function, if defined, is called by the completer when real
    filename completion is done, after all the matching names have been
    generated. It is passed a (char**) known as matches in the code below.
@@ -380,8 +391,9 @@ rl_complete_internal (what_to_do)
   else
     our_func = (Function *)filename_completion_function;
 
-  /* Only the completion entry function can change this. */
+  /* Only the completion entry function can change these. */
   rl_filename_completion_desired = 0;
+  rl_filename_quoting_desired = 1;
 
   /* We now look backwards for the start of a filename/variable word. */
   end = rl_point;
@@ -453,15 +465,20 @@ rl_complete_internal (what_to_do)
 		continue;
 #endif /* SHELL */
 
-	      /* Convoluted code, but it avoids an n2 algorithm with calls
+	      /* Convoluted code, but it avoids an n^2 algorithm with calls
 	      	 to char_is_quoted. */
 	      break;
 	    }
 	}
 
-      /* If we are at a word break, then advance past it. */
+      /* If we are at an unquoted word break, then advance past it. */
       scan = rl_line_buffer[rl_point];
+#if defined (SHELL)
+      if ((found_quote == 0 || char_is_quoted (rl_line_buffer, rl_point) == 0) &&
+          strchr (rl_completer_word_break_characters, scan))
+#else
       if (strchr (rl_completer_word_break_characters, scan))
+#endif
 	{
 	  /* If the character that caused the word break was a quoting
 	     character, then remember it as the delimiter. */
@@ -533,8 +550,12 @@ rl_complete_internal (what_to_do)
 	  /* Sort the items. */
 	  /* It is safe to sort this array, because the lowest common
 	     denominator found in matches[0] will remain in place. */
-	  for (i = 0; matches[i]; i++);
-	  qsort (matches, i, sizeof (char *), compare_strings);
+	  for (i = 0; matches[i]; i++)
+	    ;
+	  /* Try sorting the array without matches[0], since we need it to
+	     stay in place no matter what. */
+	  if (i)
+	    qsort (matches+1, i-1, sizeof (char *), compare_strings);
 
 	  /* Remember the lowest common denominator for it may be unique. */
 	  lowest_common = savestring (matches[0]);
@@ -602,12 +623,14 @@ rl_complete_internal (what_to_do)
 	  replacement = matches[0];
 
 	  should_quote = matches[0] && rl_completer_quote_characters &&
-			 rl_filename_completion_desired;
+			 rl_filename_completion_desired &&
+			 rl_filename_quoting_desired;
 
+	  if (should_quote)
 #if defined (SHELL)
-	  should_quote = should_quote && (!quote_char || quote_char == '"');
+	    should_quote = should_quote && (!quote_char || quote_char == '"');
 #else
-	  should_quote = should_quote && !quote_char;
+	    should_quote = should_quote && !quote_char;
 #endif
 
 	  if (should_quote)
@@ -831,19 +854,18 @@ rl_complete_internal (what_to_do)
 	    count = (len + (limit - 1)) / limit;
 
 	    /* Watch out for special case.  If LEN is less than LIMIT, then
-	       just do the inner printing loop. */
-	    if (len < limit)
-	      count = 1;
+	       just do the inner printing loop.
+	       0 < len <= limit  implies  count = 1. */
 
 	    /* Sort the items if they are not already sorted. */
 	    if (!rl_ignore_completion_duplicates)
-	      qsort (matches, len, sizeof (char *), compare_strings);
+	      qsort (matches + 1, len - 1, sizeof (char *), compare_strings);
 
 	    /* Print the sorted items, up-and-down alphabetically, like
 	       ls might. */
 	    crlf ();
 
-	    for (i = 1; i < count + 1; i++)
+	    for (i = 1; i <= count; i++)
 	      {
 		for (j = 0, l = i; j < limit; j++)
 		  {
@@ -944,7 +966,13 @@ static int
 compare_strings (s1, s2)
   char **s1, **s2;
 {
-  return (strcmp (*s1, *s2));
+  int result;
+
+  result = **s1 - **s2;
+  if (result == 0)
+    result = strcmp (*s1, *s2);
+
+  return result;
 }
 
 /* A completion function for usernames.
@@ -981,8 +1009,11 @@ username_completion_function (text, state)
 
   while (entry = getpwent ())
     {
-      if ((username[0] == entry->pw_name[0]) &&
-	  (strncmp (username, entry->pw_name, namelen) == 0))
+      /* Null usernames should result in all users as possible completions. */
+      if (namelen == 0)
+        break;
+      else if ((username[0] == entry->pw_name[0]) &&
+	       (strncmp (username, entry->pw_name, namelen) == 0))
 	break;
     }
 
